//
// Replicated state machine implementation with a master and several
// slaves. The master receives requests, assigns them a view stamp (a
// vid, and a sequence number) in the order of reception, and forwards
// them to all slaves. A slaves executes requests in the order of that
// the master stamps them and replies with an OK to the master. The
// master executes the request after it receives OKs from all slaves,
// and sends the reply back to the client.
//
// After a failure, and during recovery, each member of the RSM
// contacts all other members (provided by Paxos) to find who executed
// the most recent operation (based on the viewstamp).  The most
// recent one will be the new master.  Ties are broken by chosing the
// lowest numbered member.
//
// The implementation should be used only with servers that run all
// requests run to completion; in particular, a request shouldn't
// block.  If a request blocks, the slaves won't respond to the
// master, and the master won't execute the request.  A request may
// send an RPC to another host, but the RPC should be a one-way
// message to that host; the slaves shouldn't do anything based on the
// response or execute after the response, because it is not
// guaranteed that all slaves will receive the same response and
// execute in the same order.
//
// 

#include "rsm.h"
#include "rsm_client.h"
#include <fstream>
#include <iostream>
#include <signal.h>

static void *
recoverythread(void *x)
{
  rsm *r = (rsm *) x;
  r->recovery();
  return 0;
}

static bool break1 = false;

static void
breakpoint1()
{
  if (break1) {
    printf("Dying at breakpoint 1 in rsm!\n");
    exit(1);
  }
}

// This works only correctly when there's one RSM object per process
void handle_int(int param)
{
  
  if (param == SIGINT) {
    printf("Gotten SIGINT\n");
    break1 = true;
  }
}

rsm::rsm(std::string _first, std::string _me) 
  : stf(0), master(_first), stable (false)
{
  pthread_t th;


  pthread_mutex_init(&rsm_mutex, NULL);
  pthread_cond_init(&recovery_cond, NULL);

  pxs = new paxos(_first, _me, this);

  signal( SIGINT, handle_int);

  rsmrpc = pxs->get_rpcs();
  rsmrpc->reg(rsm_client_protocol::invoke, this, &rsm::client_invoke);
  rsmrpc->reg(rsm_client_protocol::members, this, &rsm::client_members);
  rsmrpc->reg(rsm_protocol::invoke, this, &rsm::invoke);
  rsmrpc->reg(rsm_protocol::transferreq, this, &rsm::transferreq);
  rsmrpc->reg(rsm_protocol::viewstampreq, this, &rsm::viewstampreq);

  //recovery thread is in charge of performing state transfer from new master
  assert(pthread_create(&th, NULL, &recoverythread, (void *) this) == 0);
}

void
rsm::reg1(int proc, handler *h)
{
  assert(pthread_mutex_lock(&rsm_mutex)==0);
  procs[proc] = h;
  assert(pthread_mutex_unlock(&rsm_mutex)==0);
}

void
rsm::recovery()
{
  struct timespec next_timeout;
  struct timeval tp;

  assert(pthread_mutex_lock(&rsm_mutex)==0);
  while(1) {
    //This node needs to perform state transfer
    //if paxos has finished performing a view change.
    //For state transfer, this node must first find the master
    //which is the node with the largest viewstamp number (with ties broken by id)
    //and then transfer state from the master.
    gettimeofday(&tp, NULL);
    next_timeout.tv_sec  = tp.tv_sec;
    next_timeout.tv_nsec = tp.tv_usec * 1000;
    next_timeout.tv_sec += 1;
    pthread_cond_timedwait(&recovery_cond, &rsm_mutex, &next_timeout);
  }
  assert(pthread_mutex_unlock(&rsm_mutex)==0);
}


/**
 * Call to transfer state from m to the local node.
 * Assumes that rsm_mutex is already held.
 */
bool
rsm::statetransfer(std::string m, viewstamp vs)
{
  std::string r;
  rpcc *which_cl = pxs->get_cl(m);
  assert(pthread_mutex_unlock(&rsm_mutex)==0);
  int ret = which_cl->call(rsm_protocol::transferreq, pxs->myaddr(), 
        vs, r, rpcc::to(2000));
  assert(pthread_mutex_lock(&rsm_mutex)==0);
  if (ret != rsm_protocol::OK)
    return false;
  if (stf) {
    stf->unmarshal_state(r);
  }
  printf("rsm::statetransfer transfer from %s success, vs(%d,%d)\n", 
   m.c_str(), vs.vid, vs.seqno);
  return true;
}

/*
 * Paxos informs rsm whenever it has successfully 
 * completed a view change
 */
void 
rsm::view_change() 
{
  pthread_mutex_lock(&rsm_mutex);
  printf("view change\n");
  stable = false;
  pthread_cond_signal(&recovery_cond);
  pthread_mutex_unlock(&rsm_mutex);
}


std::string
rsm::execute(int procno, std::string req)
{
  printf("execute\n");
  handler *h = procs[procno];
  assert(h);
  unmarshall args(req);
  marshall rep;
  std::string reps;
  rsm_protocol::status ret = h->fn(args, rep);
  marshall rep1;
  rep1 << ret;
  rep1 << rep.str();
  return rep1.str();
}


//
// Clients call client_invoke to invoke a procedure on the replicated state
// machine: the master receives the request, assigns it a sequence
// number, and invokes it on all members of the replicated state
// machine.
//
rsm_client_protocol::status
rsm::client_invoke(int procno, std::string req, std::string &r)
{
  int ret = rsm_protocol::OK;
  return ret;
}

// 
// The master calls the internal invoke at each member of the
// replicated state machine 
//
// the replica must execute requests in order (with no gaps) 
// according to requests' seqno 

rsm_protocol::status
rsm::invoke(int proc, viewstamp vs, std::string req, int &dummy)
{
  rsm_protocol::status ret = rsm_protocol::OK;
  return ret;
}

/**
 * RPC handler: Send back the local node's state to the caller
 */
rsm_protocol::status
rsm::transferreq(std::string src, viewstamp vs, std::string &r)
{
  assert(pthread_mutex_lock(&rsm_mutex)==0);
  int ret = rsm_protocol::OK;
  printf("transferreq from %s vs (%d,%d)\n", src.c_str(), vs.vid, vs.seqno);
  if (myvs != vs)
    ret = rsm_protocol::ERR;
  if (stf) 
    r = stf->marshal_state();
  assert(pthread_mutex_unlock(&rsm_mutex)==0);
  return rsm_protocol::OK;
}

/**
  * RPC handler: Send back the local node's latest viewstamp
  */
rsm_protocol::status
rsm::viewstampreq(unsigned int vid, viewstamp &r)
{
  int ret = rsm_client_protocol::OK;

  assert (pthread_mutex_lock(&rsm_mutex) == 0);
  printf("viewstampreq: my last vs (%d,%d) pxs stable %d vid %d\n", 
   last_myvs.vid, last_myvs.seqno, pxs->is_stable(), pxs->vid());
  r = last_myvs;
  assert (pthread_mutex_unlock(&rsm_mutex) == 0);
  return ret;
}

/*
 * RPC handler: Send back all the ndoes this local knows about to client
 * so the client can switch to a different master 
 * when it existing master fails
 */
rsm_client_protocol::status
rsm::client_members(int i, std::vector<std::string> &r)
{
  std::vector<std::string> m;
  assert(pthread_mutex_lock(&rsm_mutex)==0);
  pxs->get_members(m);
  m.push_back(master);
  r = m;
  printf("rsm::client_members return %s\n", pxs->print_members(r).c_str());
  assert(pthread_mutex_unlock(&rsm_mutex)==0);
  return rsm_protocol::OK;
}

std::string
rsm::get_master()
{
  if (!pxs->is_stable() || !stable)
    return "";

  return master;
}
bool
rsm::amimaster()
{
  assert(pthread_mutex_lock(&rsm_mutex)==0);
  bool r = get_master() == pxs->myaddr();
  assert(pthread_mutex_unlock(&rsm_mutex)==0);
  return r;
}

